/*
 Copyright (C) 1997 Arve Kylling

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 1, or (at your option)
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 To obtain a copy of the GNU General Public License write to the
 Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 USA.
*/

#include <stdio.h>							    
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "../src_c/ascii.h"
#include "../src_c/fortran_and_c.h"
#include "../src_c/numeric.h"
#include "../src_c/radiative.h"
#include "../src_c/sun.h"
#include "../src_c/getopt.h"
#include "../src_f/f77-uscore.h"
#include "uvspec.h"
#include "uvspec_anc.h"

int uvspec_check(input_struct Input){
  int status=0, i;
  FILE *f=NULL;

  status =0;

  /* ***FIXME*** Still need to check (also in uvspec function):
     aerosol_haze
     aerosol_scale_ssa
     aerosol_set_gg
     aerosol_set_ssa
     aerosol_season
     aerosol_visibility
     aerosol_vulcan
     angstrom
     atmosphere_file
     data_files_path
     nstr
     rte_solver
     solar_file
     sza
     sza_file
     umu
     wc_file
     wvn
     zout
   */

  /* Albedo checks */
  
  if ( Input.atm.albedo > 1.0 ) {
    fprintf (stderr, "Surface albedo > 1.0: %f\n", Input.atm.albedo);
    status++;
  }
  else if ( Input.atm.albedo < 0.0 && strlen(Input.albedo_filename)==0 ) {
    fprintf (stderr, "Surface albedo < 0.0: %f\n", Input.atm.albedo);
    status++;
  }

  /* Day of year checks */

  if ( Input.day_of_year > 366 ) {
    fprintf (stderr, "Day of year > 366: %d\n", Input.day_of_year);
    status++;
  }
  else if ( Input.day_of_year < 0 && Input.day_of_year != -999 ) {
    fprintf (stderr, "Day of year < 0: %d\n", Input.day_of_year);
    status++;
  }

  /* Ozone column checks */

  if ( Input.ozone_column < 0.0 ) {
    fprintf (stderr, "Ozone_column < 0.0: %f\n", Input.ozone_column);
    status++;
  }

  /* RTE checks */

  if ( (Input.rte.solver==SDISORT) && strlen(Input.watercloud_filename)>0 ) {
    fprintf (stderr, "Warning: sdisort is not suited for calculations involving clouds\n");
  }
  else if ( (Input.rte.solver==TWOSTR) && Input.rte.nprndis>0 ) {
    for (i=0;i<Input.rte.nprndis;i++) {
      if ( !(Input.rte.prndis[i] == 1 || Input.rte.prndis[i] == 2) ) { 
	fprintf (stderr, "Twostr can only handle prndis set to 1 and/or 2\n");
	status++;
      }
      else {
	fclose(f);
      }
    }
  }

  /* Check wavelength ranges */
  if ( Input.spline ) {
    if ( Input.wl.wl_start > Input.spline_lambda_0 )  {
      fprintf (stderr, "Spline lambda %f is smaller than wvn lambda %f\n",
	       Input.spline_lambda_0, Input.wl.wl_start);
      status++;
    }
    if ( Input.wl.wl_end < Input.spline_lambda_1 )  {
      fprintf (stderr, "Spline lambda %f is greater than wvn lambda %f\n",
	       Input.spline_lambda_1, Input.wl.wl_end);
      status++;
    }
  }

  /* Check that slit_function file exists */
  if ( Input.convolve ) {
    if ( (f = fopen(Input.slit_function_filename, "r")) == NULL) {
      fprintf (stderr, "Can not read slit function file %s\n", Input.slit_function_filename);
      status++;
    }
  }

  return status;
}

int uvspec(input_struct input, output_struct *output)
{
  int iu, iv, j, lc, lev, lu;
  int linear = 0;
  int status = 0;
  int maxawvn = MAXAWVN;
  int rows=0, min_columns=0, max_columns=0;
  float *tmp_aer_dtau, *tmp_aer_gg, *tmp_aer_ssa, *tmp_dens, 
    *tmp_ice_dtau, *tmp_ice_gg, *tmp_ice_ssa,
    *tmp_wc_dtau, *tmp_wc_gg, *tmp_wc_ssa,
    *tmp_aerext, *tmp_aerabs, *tmp_aersym, *tmp_aerwvn;  
  float *tmp_dfdt, *tmp_flup, *tmp_rfldir, *tmp_rfldn, *tmp_uavg, *tmp_uavgdn,
    *tmp_uavgso, *tmp_uavgup, *tmp_u0u, *tmp_uu;
  float *tmp_dfdt_t, *tmp_flup_t, *tmp_rfldir_t, *tmp_rfldn_t, *tmp_uavg_t, *tmp_uavgdn_t,
    *tmp_uavgso_t, *tmp_uavgup_t, *tmp_u0u_t, *tmp_albmed, *tmp_trnmed, *tmp_uu_t,
    *tmp_utau, *tmp_phi;
  float *tmp_temper, *tmp_zd;
  float **wc_data=NULL, **wl_data=NULL, sum, scale_factor;

  void F77_FCN (aerprof) (int* nlyr, float* zd, float* aer_pro, 
			   int* seasn, int* vulcan, float* visib );
  void F77_FCN (aeropt) ( int* vulcan, float* aerext, float* aerabs, 
			  float* aersym, float* aerwvn, float* rh, 
			  int* nlyr, float* zd, int* haze, int* maxawvn ); 
  void F77_FCN (aerwvn) ( float* aerwvn, float* aerabs, float* aersym, float* aerext,
			  float* aer_pro, float* aer_dtau, float* aer_gg, 
			  float* aer_ssa, int* nlyr, float* lambda_t, float* zd, 
			  int* nlambda_t, int* maxawvn );

  float F77_FCN (o3column) (float* dens, int* nlyr, float* zd);
  void  F77_FCN (scaleaer) (float* aer_dtau, float* lambda_t, int* nlambda_t, 
			    float* alpha, float* beta, int* nlyr);
  void  F77_FCN (scaleozone) (float* dens, int* nlyr, 
			       float* ozone_column, float* zd, float* ozone_scale_factor);
  void  F77_FCN (trans) (float* aer_dtau, float* aer_gg, float* aer_ssa, 
			 float* albedo_t, float* dens,  
			 float* ice_dtau, float* ice_gg, float* ice_ssa, 
			 float* lambda_t, int* ndens, int* nlambda_t, int* nlyr, 
			 int* nzout, int* maxphi, int* maxumu, int* nstr, int* numu, int* nphi,
			 const char* filepath, int* nstring,
			 float* press, int* solver, int* deltam, int* o3_type,
			 float* sza_t, float* phi0_t, float* temper, int* fisot, float* umu, 
			 int* nprndis, int* prndis, int* reverse,
			 float* wc_dtau, float* wc_gg, 
			 float* wc_ssa, float* zd, float* zout, float* phi, float* utau,
			 float* dfdt, float* flup, float* rfldir, float* rfldn, 
			 float* uavg,  float* uavgdn, float* uavgso, float* uavgup,
			 float* u0u, float* albmed, float* trnmed, float* uu);
  void  F77_FCN (transtest) (float* aer_dtau, int* umaxcly); 
  void  F77_FCN (wcloud) (float* lambda_t, int* newsiz, int* nlyr,
			  char* filepath, int* nstring, float* wccon, float* wceffr, 
			  float* tmp_wc_dtau, float* tmp_wc_gg, float* tmp_wc_ssa, float* zd );
    
  int nstring;
  nstring = (int)strlen(input.filepath);
  
  /**** Calculate cmu (computational quadrature angles) if user requests it ****/
  
  
  /**** Set wavelength grid for the transmittance calculation ****/
  
  if ( strlen(input.transmittance_wl_file) > 0 ) {
    
    status = ASCII_file2float (input.transmittance_wl_file, 
			       &rows, &max_columns, &min_columns, 
			       &wl_data);
    if (status!=0) {
      fprintf (stderr, " ... error %d opening file %s\n", status, input.transmittance_wl_file);
      return status;
    }
    if (max_columns!=min_columns) {
      fprintf (stderr, " !! ATTENTION !! Inconsistent number of columns\n");
      fprintf (stderr, "     min = %d, max =%d\n", min_columns, max_columns);
    }  
    if (min_columns!=1) {
      fprintf (stderr, " ... ending, too few columns\n");
      return -1;
    }
    output->wl.nlambda_t = rows;
    output->wl.lambda_t  = ASCII_column_float(wl_data, output->wl.nlambda_t, 0 );
  }
  else {
    status = set_transmittance_wl_grid(&input.wl, &output->wl);
  }
  
  /**** Read in model atmosphere ****/
  
  status = read_atmosphere(input.atmosphere_filename, 
			   &input.atm, &output->atm, input.crs.column);
  if (status!=0) {
    fprintf (stderr, " ... error %d during handling of file %s\n",
	     status, input.atmosphere_filename);
    return status;
  }

  output->atm.rh  = (float *) calloc (output->atm.nlev, sizeof(float));

  /* Shift atmosphere to given altitude */
  if (input.atm.altitude != 0) {
    status = shift_atmosphere(&input.atm, &output->atm);
    if (status !=0) {
      fprintf (stderr, "error %d shifting atmosphere\n", status);
      return status;
    }
  }

  /* Linearly interpolate output temperature to zout */
  /* sort profile in ascending order, assuming that */
  /* the original profile was descending            */
  tmp_zd     = (float *) calloc(output->atm.nlev, sizeof(float));
  tmp_temper = (float *) calloc(output->atm.nlev, sizeof(float));
  for (lev=0; lev<output->atm.nlev; lev++) {
    tmp_zd[lev]       = output->atm.zd          [output->atm.nlev-1-lev];
    tmp_temper[lev]   = output->atm.temper      [output->atm.nlev-1-lev];
  }
  output->atm.temper_out  = (float *) calloc (input.atm.nzout, sizeof(float));
  linear = 1;
  status = arb_wvn( output->atm.nlev, tmp_zd, tmp_temper, 
			input.atm.nzout, input.atm.zout, output->atm.temper_out, linear);
  if ( status != 0.0 ) {
    fprintf (stderr, "Error interpolating temperature to output grid\n");
    return status;
  }
  free(tmp_temper); free(tmp_zd);

  input.rte.nlev = output->atm.nlev;
  input.rte.nlyr = output->atm.nlyr;

  /**** Read extra cross section files, like BrO, OclO, NO2, etc. ****/
  /**** FIXME: YET TO BE IMPLEMENTED   ****/

  if ( strlen(input.crs.filename) > 0 ) {
    printf("crs: %d %s\n",input.crs.column,input.crs.filename);
    return -1;
  }

  /**** Scale pressure and air profiles ****/
  if ( input.pressure >= 0.0 ) {
    scale_factor = input.pressure/output->atm.press[output->atm.nlyr];
    for (lc=0;lc<=output->atm.nlyr;lc++) {
      output->atm.press[lc] *= scale_factor;
      output->atm.dens[AIR][lc] *= scale_factor;
    } 
  }

  /**** Scale ozone column ****/

  output->ozone_column = F77_FCN (o3column) ( output->atm.dens[O3], &input.rte.nlyr, output->atm.zd );

  if ( input.ozone_column >= 0.0 ) {
    F77_FCN (scaleozone) ( output->atm.dens[O3], &output->atm.nlyr, &input.ozone_column,
			   output->atm.zd, &output->ozone_scale_factor );
    output->ozone_column = F77_FCN (o3column) ( output->atm.dens[O3], &output->atm.nlyr, output->atm.zd ); 
  }
  
  /****  Read extraterrestrial spectrum ****/

  status = read_extraterrestrial(input.extraterrestrial_filename, 
				 &input.wl, &output->wl, 
				 input.wl.wl_start, input.wl.wl_end);
  if (status!=0) {
    fprintf (stderr, " ... error %d during handling of file %s\n",
	     status, input.extraterrestrial_filename);
    return status;
  }

  /**** Correction factor for Earth-Sun distance ****/

  output->eccentricity_factor = 1.0;

  if ( input.day_of_year > 0.0 ) {
    output->eccentricity_factor = eccentricity( input.day_of_year );
    for (iv=0;iv<output->wl.nlambda_h;iv++) {
      output->wl.fbeam[iv] *= output->eccentricity_factor;
    }
  }

  /**** Some memory allocation ****/

  if ( (status = ASCII_calloc_float (&input.ice.dtau, output->wl.nlambda_t, input.rte.nlyr)) != 0 ) return status;
  if ( (status = ASCII_calloc_float (&input.ice.gg,   output->wl.nlambda_t, input.rte.nlyr)) != 0 ) return status;
  if ( (status = ASCII_calloc_float (&input.ice.ssa,  output->wl.nlambda_t, input.rte.nlyr)) != 0 ) return status;

  output->atm.albedo_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));
  output->atm.sza_t    = (float *) calloc (output->wl.nlambda_t, sizeof(float));
  output->atm.phi0_t   = (float *) calloc (output->wl.nlambda_t, sizeof(float));

  /**** Allow user to specify wavelength dependent albedo in seperate file ****/

  if ( input.atm.albedo < 0.0 ) {
    status = read_albedo(input.albedo_filename, &input.atm, &output->atm);
    if ( status != 0.0 ) {
      fprintf (stderr, "Error reading albedo file: %s\n", input.albedo_filename);
      return status;
    }

    /**** Interpolate albedo to transmission grid ****/
    
    linear = 1; /**** Use linear interpolation for the albedo, otherwise 
                      negative albedos or albedos larger than 1 may occurr, 
                      I should know:-) AKy, ****/ 
    status = arb_wvn( output->atm.file_nlambda_albedo, output->atm.file_lambda_albedo,
		      output->atm.file_albedo, 
		      output->wl.nlambda_t, output->wl.lambda_t, output->atm.albedo_t, linear);
    if ( status != 0.0 ) {
      fprintf (stderr, "arb_wvn reports error %d\n", 
	       status);
      return status;
    }

    /**** Check interpolated albedo values ****/

    for (iv=0;iv<output->wl.nlambda_t;iv++) {
      if ( output->atm.albedo_t[iv] > 1.0 ){
	fprintf (stderr, "Albedo interpolated from file is > 1.0. Wavelength: %f, albedo %f\n", 
		 output->wl.lambda_t[iv], output->atm.albedo_t[iv] );
	fprintf (stderr, "Modify your albedo file: %s\n",input.albedo_filename);
	status=1;
	return status;
      } 
      else 
      if ( output->atm.albedo_t[iv] < 0.0 ){
	fprintf (stderr, "Albedo interpolated from file is < 0.0: %f for wavelength %6.2f\n", 
		 output->atm.albedo_t[iv],output->wl.lambda_t[iv]);
	fprintf (stderr, "Modify your albedo file: %s\n",input.albedo_filename);
	status=2;
	return status;
      } 
    }
  }
  else {
    for (iv=0;iv<output->wl.nlambda_t;iv++) {
      output->atm.albedo_t[iv] = input.atm.albedo;    
    }
  }

  /**** Allow user to specify sza as a function of wavelength ****/

  if ( input.atm.sza < -999.0 ) {
    status = read_sza(input.sza_filename, &input.atm, &output->atm);
    if ( status != 0.0 ) {
      fprintf (stderr, "Error reading sza file: %s\n",
	       input.sza_filename);
      return status;
    }

    /**** Interpolate sza to transmission grid ****/

    linear = 0;
    status = arb_wvn( output->atm.file_nlambda_sza, output->atm.file_lambda_sza,
		      output->atm.file_sza, 
		      output->wl.nlambda_t, output->wl.lambda_t, output->atm.sza_t, linear);
    status = arb_wvn( output->atm.file_nlambda_sza, output->atm.file_lambda_sza,
		      output->atm.file_phi0, 
		      output->wl.nlambda_t, output->wl.lambda_t, output->atm.phi0_t, linear);
  }
  else {
    for (iv=0;iv<output->wl.nlambda_t;iv++) {
      output->atm.sza_t[iv]  = input.atm.sza;    
      output->atm.phi0_t[iv] = input.atm.phi0;    
    }
  }



  /**** Water cloud stuff ****/

  if ( strlen(input.watercloud_filename) > 0 ) {

    status = ASCII_file2float (input.watercloud_filename, 
			      &rows, &max_columns, &min_columns, 
			      &wc_data);
    if (status!=0) {
      fprintf (stderr, " ... error %d opening file %s\n", status, input.watercloud_filename);
      return status;
    }
    if (max_columns!=min_columns) {
      fprintf (stderr, " !! ATTENTION !! Inconsistent number of columns\n");
      fprintf (stderr, "     min = %d, max =%d\n", min_columns, max_columns);
    }  
    if (min_columns<2) {
      fprintf (stderr, " ... ending, too few columns\n");
      return -1;
    }
    output->wc.wcnlev = rows;
    if ( output->wc.wcnlev != output->atm.nlev ) {
      fprintf (stderr, "Water cloud layers differes from atmosphere layers\n");      
      return -1;
    }
    output->wc.wczd   = ASCII_column_float(wc_data, output->atm.nlev, WCZD );
    for (lev=0;lev<output->wc.wcnlev;lev++) {
      if ( output->wc.wczd[lev] != output->atm.zd[lev] ) {
	fprintf (stderr, "Water cloud layers differes from atmosphere layers %d %f %f\n", 
		 lev, output->wc.wczd[lev], output->atm.zd[lev]);      
	return -1;
      }
    }
    output->wc.wccon  = ASCII_column_float(wc_data, output->atm.nlev, WCCON );
    output->wc.wceffr = ASCII_column_float(wc_data, output->atm.nlev, WCEFFR );

    output->wc.newsiz = 1;

    tmp_wc_dtau       = (float *) calloc (input.rte.nlyr, sizeof(float));
    tmp_wc_gg         = (float *) calloc (input.rte.nlyr, sizeof(float));
    tmp_wc_ssa        = (float *) calloc (input.rte.nlyr, sizeof(float));
    
    for (iv=0;iv<output->wl.nlambda_t;iv++) {
      F77_FCN (wcloud) ( &output->wl.lambda_t[iv], &output->wc.newsiz, &output->atm.nlyr,
			 input.filepath, &nstring, output->wc.wccon, output->wc.wceffr, 
			 tmp_wc_dtau, tmp_wc_gg, tmp_wc_ssa, output->atm.zd );
      if (iv == 0) {
	if ( (status = ASCII_calloc_float (&output->wc.dtau,  output->wl.nlambda_t, input.rte.nlyr)) != 0 ) return status;
	if ( (status = ASCII_calloc_float (&output->wc.gg,    output->wl.nlambda_t, input.rte.nlyr)) != 0 ) return status;
	if ( (status = ASCII_calloc_float (&output->wc.ssa,   output->wl.nlambda_t, input.rte.nlyr)) != 0 ) return status;
      } 

      /**** Set the water cloud optical depth to a user specified value ****/
      if ( input.wc.wctau >= 0.0 ) {
	sum = 0.0;
	for (lev=0;lev<output->atm.nlyr;lev++) {
	  sum += tmp_wc_dtau[lev];
	}
	scale_factor = input.wc.wctau/sum;
	for (lev=0;lev<output->atm.nlyr;lev++) {
	  tmp_wc_dtau[lev] = tmp_wc_dtau[lev]*scale_factor;
	}
      }

      for (lev=0;lev<output->atm.nlyr;lev++) {
	output->wc.dtau[iv][lev] = tmp_wc_dtau[lev];
	output->wc.gg[iv][lev]   = tmp_wc_gg[lev];
	output->wc.ssa[iv][lev]  = tmp_wc_ssa[lev];
      }
    }
    free(tmp_wc_dtau); 
    free(tmp_wc_gg); 
    free(tmp_wc_ssa); 
    status = ASCII_free_float(wc_data, rows);
  }
  else {
    if ( (status = ASCII_calloc_float (&output->wc.dtau,  output->wl.nlambda_t, input.rte.nlyr)) != 0 ) return status;
    if ( (status = ASCII_calloc_float (&output->wc.gg,    output->wl.nlambda_t, input.rte.nlyr)) != 0 ) return status;
    if ( (status = ASCII_calloc_float (&output->wc.ssa,   output->wl.nlambda_t, input.rte.nlyr)) != 0 ) return status;
  }

  /****   Aerosol stuff ****/
  if ( input.aer.vulcan > 0 ) {

    /****   Get appropriate aerosol profile and interpolate profile to zd grid.****/

    output->aer.aer_pro  = (float *) calloc(output->atm.nlev, sizeof(float));
    F77_FCN (aerprof) ( &output->atm.nlyr, output->atm.zd, output->aer.aer_pro, 
			&input.aer.seasn, &input.aer.vulcan, &input.aer.visib );

    /****   Get appropriate aerosol optical properties. ****/

    tmp_aerext       = (float *) calloc (maxawvn*output->atm.nlev, sizeof(float));
    tmp_aerabs       = (float *) calloc (maxawvn*output->atm.nlev, sizeof(float));
    tmp_aersym       = (float *) calloc (maxawvn*output->atm.nlev, sizeof(float));
    tmp_aerwvn       = (float *) calloc (maxawvn*output->atm.nlev, sizeof(float));
    F77_FCN (aeropt) ( &input.aer.vulcan, tmp_aerext, tmp_aerabs, 
		       tmp_aersym, tmp_aerwvn, output->atm.rh, 
		       &output->atm.nlyr, output->atm.zd, &input.aer.haze, &maxawvn ); 

    tmp_aer_dtau       = (float *) calloc (output->wl.nlambda_t*input.rte.nlyr, sizeof(float));
    tmp_aer_gg         = (float *) calloc (output->wl.nlambda_t*input.rte.nlyr, sizeof(float));
    tmp_aer_ssa        = (float *) calloc (output->wl.nlambda_t*input.rte.nlyr, sizeof(float));
    F77_FCN (aerwvn) ( tmp_aerwvn, tmp_aerabs, tmp_aersym, tmp_aerext, output->aer.aer_pro, 
			tmp_aer_dtau, tmp_aer_gg, tmp_aer_ssa, &output->atm.nlyr,
			output->wl.lambda_t, output->atm.zd, &output->wl.nlambda_t, &maxawvn );

    output->aer.gg   = fortran2c_2D_float_ary(output->wl.nlambda_t, input.rte.nlyr, tmp_aer_gg );
    output->aer.ssa  = fortran2c_2D_float_ary(output->wl.nlambda_t, input.rte.nlyr, tmp_aer_ssa );
    output->aer.dtau = fortran2c_2D_float_ary(output->wl.nlambda_t, input.rte.nlyr, tmp_aer_dtau );
    
    free(tmp_aer_gg);
    free(tmp_aer_ssa);
    free(tmp_aer_dtau);

    /****   Read in aerosol optical depth file if specified ****/
    if ( strlen(input.aer.tau_filename) > 0 ) {
      status = read_aerosol_tau(input.aer.tau_filename, input.rte.nlyr,
				output->wl.nlambda_t, output->atm.zd,
				output->aer.dtau);
      if ( status != 0.0 ) {
	fprintf (stderr, "Error reading aerosol tau file: %s\n", 
		 input.aer.tau_filename);
	return status;
      }
    }

    /**** Set aerosol asymmetry factor ****/
    if (input.aer.set_gg >= -1.0 ) {
      for (lc=0;lc<input.rte.nlyr;lc++) {
	for (iv=0;iv<output->wl.nlambda_t;iv++) {
	  /*printf("set: %d %d %f %f\n",lc,iv,output->atm.zd[lc],output->aer.ssa[iv][lc]);*/
	  output->aer.gg[iv][lc]  = input.aer.set_gg;
	}
      }
    }
    /**** Set aerosol single scattering albedo ****/
    if (input.aer.set_ssa >= 0.0 ) {
      for (lc=0;lc<input.rte.nlyr;lc++) {
	for (iv=0;iv<output->wl.nlambda_t;iv++) {
	  /*printf("set: %d %d %f %f\n",lc,iv,output->atm.zd[lc],output->aer.ssa[iv][lc]);*/
	  output->aer.ssa[iv][lc]  = input.aer.set_ssa;
	}
      }
    }
    /**** Scale aerosol single scattering albedo ****/
    if (input.aer.scale_ssa > 0.0 ) {
      for (lc=0;lc<input.rte.nlyr;lc++) {
	for (iv=0;iv<output->wl.nlambda_t;iv++) {
	  output->aer.ssa[iv][lc]  = output->aer.ssa[iv][lc] * input.aer.scale_ssa;
	  if (output->aer.ssa[iv][lc]>1.0) {output->aer.ssa[iv][lc] = 1.0;}
	}
      }
    }
    free(tmp_aerext);
    free(tmp_aerabs);
    free(tmp_aersym);
    free(tmp_aerwvn);

    if ( input.aer.beta >= 0.0 ) {
      tmp_aer_dtau  = c2fortran_2D_float_ary(output->wl.nlambda_t, 
					     input.rte.nlyr, output->aer.dtau );
      F77_FCN (scaleaer) (tmp_aer_dtau, output->wl.lambda_t, &output->wl.nlambda_t, 
			  &input.aer.alpha, &input.aer.beta, &input.rte.nlyr);
      output->aer.dtau = fortran2c_2D_float_ary(output->wl.nlambda_t, 
						input.rte.nlyr, tmp_aer_dtau );
      free(tmp_aer_dtau);
    }

  }
  else {
    if ((status = ASCII_calloc_float(&output->aer.dtau, output->wl.nlambda_t, input.rte.nlyr)) != 0 ) return status;
    if ((status = ASCII_calloc_float(&output->aer.gg,   output->wl.nlambda_t, input.rte.nlyr)) != 0 ) return status;
    if ((status = ASCII_calloc_float(&output->aer.ssa,  output->wl.nlambda_t, input.rte.nlyr)) != 0 ) return status;
  }
  
  /**** Move all 2D C arys into 1D arys compatible with fortran, do it both for input and output
     arys. For the output arys only a large enough ary is required, no initialization needed. ****/

  tmp_aer_dtau   = c2fortran_2D_float_ary(output->wl.nlambda_t, input.rte.nlyr, output->aer.dtau );
  tmp_aer_gg     = c2fortran_2D_float_ary(output->wl.nlambda_t, input.rte.nlyr, output->aer.gg   );
  tmp_aer_ssa    = c2fortran_2D_float_ary(output->wl.nlambda_t, input.rte.nlyr, output->aer.ssa  );
  tmp_ice_dtau   = c2fortran_2D_float_ary(output->wl.nlambda_t, input.rte.nlyr, input.ice.dtau );
  tmp_ice_gg     = c2fortran_2D_float_ary(output->wl.nlambda_t, input.rte.nlyr, input.ice.gg   );
  tmp_ice_ssa    = c2fortran_2D_float_ary(output->wl.nlambda_t, input.rte.nlyr, input.ice.ssa  );
  tmp_wc_dtau    = c2fortran_2D_float_ary(output->wl.nlambda_t, input.rte.nlyr, output->wc.dtau );
  tmp_wc_gg      = c2fortran_2D_float_ary(output->wl.nlambda_t, input.rte.nlyr, output->wc.gg   );
  tmp_wc_ssa     = c2fortran_2D_float_ary(output->wl.nlambda_t, input.rte.nlyr, output->wc.ssa  );
  tmp_dens       = c2fortran_2D_float_ary(output->atm.ndens,  output->atm.nlev, output->atm.dens );

  tmp_albmed       = (float *) calloc (input.rte.maxumu, sizeof(float));
  tmp_trnmed       = (float *) calloc (input.rte.maxumu, sizeof(float));
  tmp_phi          = (float *) calloc (input.rte.maxphi, sizeof(float));
  tmp_utau         = (float *) calloc (input.atm.nzout, sizeof(float));
  tmp_rfldn_t      = (float *) calloc (output->wl.nlambda_t*input.atm.nzout, sizeof(float));
  tmp_uavg_t       = (float *) calloc (output->wl.nlambda_t*input.atm.nzout, sizeof(float));
  tmp_uavgso_t     = (float *) calloc (output->wl.nlambda_t*input.atm.nzout, sizeof(float));
  tmp_rfldir_t     = (float *) calloc (output->wl.nlambda_t*input.atm.nzout, sizeof(float));
  tmp_dfdt_t       = (float *) calloc (output->wl.nlambda_t*input.atm.nzout, sizeof(float));
  tmp_flup_t       = (float *) calloc (output->wl.nlambda_t*input.atm.nzout, sizeof(float));
  tmp_uavgdn_t     = (float *) calloc (output->wl.nlambda_t*input.atm.nzout, sizeof(float));
  tmp_uavgup_t     = (float *) calloc (output->wl.nlambda_t*input.atm.nzout, sizeof(float));
  tmp_u0u_t        = (float *) calloc (output->wl.nlambda_t*input.atm.nzout*input.rte.maxumu, sizeof(float));
  tmp_uu_t         = (float *) calloc (output->wl.nlambda_t*input.atm.nzout*input.rte.maxumu*input.rte.maxphi, sizeof(float));

  /**** Calculate radiation field with fbeam = 1.0, no extraterrestrial 
        spectrum involved. ****/
  
  switch (input.rte.solver) {
  case DISORT:
  case SDISORT:
  case TWOSTR:
    F77_FCN  (trans) (tmp_aer_dtau, tmp_aer_gg, tmp_aer_ssa, output->atm.albedo_t, 
		      tmp_dens, tmp_ice_dtau, tmp_ice_gg, tmp_ice_ssa, 
		      output->wl.lambda_t, &output->atm.ndens, &output->wl.nlambda_t,
		      &output->atm.nlyr, &input.atm.nzout, &input.rte.maxphi, 
		      &input.rte.maxumu, &input.rte.nstr, &input.rte.numu, &input.rte.nphi,
		      input.filepath, &nstring, output->atm.press, &input.rte.solver,
		      &input.rte.deltam, &input.crs.o3_type,
		      output->atm.sza_t, output->atm.phi0_t, output->atm.temper, &input.rte.fisot,
		      input.rte.umu, &input.rte.nprndis, input.rte.prndis, &input.rte.reverse,
		      tmp_wc_dtau, tmp_wc_gg, tmp_wc_ssa, output->atm.zd, input.atm.zout,
		      input.rte.phi, tmp_utau,
		      tmp_dfdt_t, tmp_flup_t, tmp_rfldir_t, tmp_rfldn_t, tmp_uavg_t,  
		      tmp_uavgdn_t, tmp_uavgso_t, tmp_uavgup_t, tmp_u0u_t,
		      tmp_albmed, tmp_trnmed, tmp_uu_t);
    break;
  case SUCORD:
    /**** FIXME NOT YET IMPLEMENTED ****/
    /**** crs_o3 ****/
    output->crs.ray = crs_rayleigh(output->wl.lambda_t, output->wl.nlambda_t);
    for (iv=0;iv<output->wl.nlambda_t;iv++) {
      printf("crs %d %e\n", iv, output->crs.ray[iv]);
      optical_properties(&input.atm, &output->atm, &output->crs);
      /*
	set_utau():
	rte_solver;
	*/
    }
    sucordsca();
    break;
  }

  free(tmp_aer_dtau); 
  free(tmp_aer_gg); 
  free(tmp_aer_ssa); 
  free(tmp_dens); 
  free(tmp_ice_dtau); 
  free(tmp_ice_gg); 
  free(tmp_ice_ssa); 
  free(tmp_wc_dtau); 
  free(tmp_wc_gg); 
  free(tmp_wc_ssa); 
  free(tmp_albmed); 
  free(tmp_trnmed); 
  
  /**** Move all output 1D fortran compatible arys into 2D C arys. Memory allocation for
     all output arys done inside fortran2c_2D_float. ****/
  
  output->rfldir_t = fortran2c_2D_float_ary(output->wl.nlambda_t, input.atm.nzout, tmp_rfldir_t );
  output->rfldn_t  = fortran2c_2D_float_ary(output->wl.nlambda_t, input.atm.nzout, tmp_rfldn_t );
  output->flup_t   = fortran2c_2D_float_ary(output->wl.nlambda_t, input.atm.nzout, tmp_flup_t );
  output->dfdt_t   = fortran2c_2D_float_ary(output->wl.nlambda_t, input.atm.nzout, tmp_dfdt_t );
  output->uavg_t   = fortran2c_2D_float_ary(output->wl.nlambda_t, input.atm.nzout, tmp_uavg_t );
  output->uavgdn_t = fortran2c_2D_float_ary(output->wl.nlambda_t, input.atm.nzout, tmp_uavgdn_t );
  output->uavgso_t = fortran2c_2D_float_ary(output->wl.nlambda_t, input.atm.nzout, tmp_uavgso_t );
  output->uavgup_t = fortran2c_2D_float_ary(output->wl.nlambda_t, input.atm.nzout, tmp_uavgup_t );
  output->u0u_t    = fortran2c_3D_float_ary(output->wl.nlambda_t, input.atm.nzout, input.rte.maxumu, tmp_u0u_t );
  output->uu_t     = fortran2c_4D_float_ary(output->wl.nlambda_t, input.atm.nzout, input.rte.maxphi, input.rte.maxumu ,tmp_uu_t );
   
  /**** Allocate memory for final output ****/
  
  tmp_flup         = (float *) calloc (output->wl.nlambda_h, sizeof(float));
  tmp_rfldir       = (float *) calloc (output->wl.nlambda_h, sizeof(float));
  tmp_rfldn        = (float *) calloc (output->wl.nlambda_h, sizeof(float));
  tmp_uavg         = (float *) calloc (output->wl.nlambda_h, sizeof(float));
  tmp_uavgdn       = (float *) calloc (output->wl.nlambda_h, sizeof(float));
  tmp_uavgso       = (float *) calloc (output->wl.nlambda_h, sizeof(float));
  tmp_uavgup       = (float *) calloc (output->wl.nlambda_h, sizeof(float));
  tmp_u0u          = (float *) calloc (output->wl.nlambda_h, sizeof(float));
  tmp_uu           = (float *) calloc (output->wl.nlambda_h, sizeof(float));

  if ( (status = ASCII_calloc_float (&output->flup,   output->wl.nlambda_h, input.atm.nzout)) != 0 ) return status;
  if ( (status = ASCII_calloc_float (&output->rfldn,  output->wl.nlambda_h, input.atm.nzout)) != 0 ) return status;
  if ( (status = ASCII_calloc_float (&output->rfldir, output->wl.nlambda_h, input.atm.nzout)) != 0 ) return status;
  if ( (status = ASCII_calloc_float (&output->uavg,   output->wl.nlambda_h, input.atm.nzout)) != 0 ) return status;
  if ( (status = ASCII_calloc_float (&output->uavgdn, output->wl.nlambda_h, input.atm.nzout)) != 0 ) return status;
  if ( (status = ASCII_calloc_float (&output->uavgso, output->wl.nlambda_h, input.atm.nzout)) != 0 ) return status;
  if ( (status = ASCII_calloc_float (&output->uavgup, output->wl.nlambda_h, input.atm.nzout)) != 0 ) return status;
  if ( (status = ASCII_calloc_float_3D (&output->u0u,    output->wl.nlambda_h, 
				     input.atm.nzout, input.rte.maxumu )) != 0 ) return status;
  if ( (status = ASCII_calloc_float_4D (&output->uu,    output->wl.nlambda_h, input.atm.nzout, 
					input.rte.maxphi, input.rte.maxumu)) != 0 ) return status;

  /**** Interpolate to output wavelength grid for each output altitude ****/

  output->sza = (float *) calloc (output->wl.nlambda_h, sizeof(float));
  for (lu=0;lu<input.atm.nzout;lu++) {
    for (iv=0;iv<output->wl.nlambda_t;iv++) {
      tmp_flup_t[iv]   = output->flup_t[iv][lu];
      tmp_rfldn_t[iv]  = output->rfldn_t[iv][lu];
      tmp_rfldir_t[iv] = output->rfldir_t[iv][lu];
      tmp_uavg_t[iv]   = output->uavg_t[iv][lu];
      tmp_uavgdn_t[iv] = output->uavgdn_t[iv][lu];
      tmp_uavgso_t[iv] = output->uavgso_t[iv][lu];
      tmp_uavgup_t[iv] = output->uavgup_t[iv][lu];
    }

    /**** No need to interpolate if only one wavelength is wanted ****/
    if ( output->wl.nlambda_t == 1 ) {
      iv = 0;
      tmp_flup[iv]      = tmp_flup_t[iv];
      tmp_rfldir[iv]    = tmp_rfldir_t[iv];
      tmp_rfldn[iv]     = tmp_rfldn_t[iv];
      tmp_uavg[iv]      = tmp_uavg_t[iv];
      tmp_uavgdn[iv]    = tmp_uavgdn_t[iv];
      tmp_uavgso[iv]    = tmp_uavgso_t[iv];
      tmp_uavgup[iv]    = tmp_uavgup_t[iv];
      output->sza[iv]   = output->atm.sza_t[iv];
    } 
    else {
      linear = 0;
      status = arb_wvn( output->wl.nlambda_t, output->wl.lambda_t, tmp_flup_t, 
			output->wl.nlambda_h, output->wl.lambda_h, tmp_flup, linear);
      status = arb_wvn( output->wl.nlambda_t, output->wl.lambda_t, tmp_rfldir_t, 
			output->wl.nlambda_h, output->wl.lambda_h, tmp_rfldir, linear);
      status = arb_wvn( output->wl.nlambda_t, output->wl.lambda_t, tmp_rfldn_t, 
			output->wl.nlambda_h, output->wl.lambda_h, tmp_rfldn, linear);
      status = arb_wvn( output->wl.nlambda_t, output->wl.lambda_t, tmp_uavg_t, 
			output->wl.nlambda_h, output->wl.lambda_h, tmp_uavg, linear);
      status = arb_wvn( output->wl.nlambda_t, output->wl.lambda_t, tmp_uavgdn_t, 
			output->wl.nlambda_h, output->wl.lambda_h, tmp_uavgdn, linear);
      status = arb_wvn( output->wl.nlambda_t, output->wl.lambda_t, tmp_uavgso_t, 
			output->wl.nlambda_h, output->wl.lambda_h, tmp_uavgso, linear);
      status = arb_wvn( output->wl.nlambda_t, output->wl.lambda_t, tmp_uavgup_t, 
			output->wl.nlambda_h, output->wl.lambda_h, tmp_uavgup, linear);
      status = arb_wvn( output->wl.nlambda_t, output->wl.lambda_t, output->atm.sza_t, 
			output->wl.nlambda_h, output->wl.lambda_h, output->sza, linear);
    }

    for (iv=0;iv<output->wl.nlambda_h;iv++) {
      output->flup[iv][lu]   = tmp_flup[iv]*output->wl.fbeam[iv];
      output->rfldn[iv][lu]  = tmp_rfldn[iv]*output->wl.fbeam[iv];
      output->rfldir[iv][lu] = tmp_rfldir[iv]*output->wl.fbeam[iv];
      output->uavg[iv][lu]   = tmp_uavg[iv]*output->wl.fbeam[iv];
      output->uavgdn[iv][lu] = tmp_uavgdn[iv]*output->wl.fbeam[iv];
      output->uavgso[iv][lu] = tmp_uavgso[iv]*output->wl.fbeam[iv];
      output->uavgup[iv][lu] = tmp_uavgup[iv]*output->wl.fbeam[iv];
    }

    for (iu=0;iu<input.rte.maxumu;iu++) {
      for (iv=0;iv<output->wl.nlambda_t;iv++) {
	tmp_u0u_t[iv] = output->u0u_t[iv][lu][iu];
      }
      if ( output->wl.nlambda_t == 1 ) {
	iv = 0;
	tmp_u0u[iv]      = tmp_u0u_t[iv];
      } 
      else {
	linear = 0;
	status = arb_wvn( output->wl.nlambda_t, output->wl.lambda_t, tmp_u0u_t, 
			  output->wl.nlambda_h, output->wl.lambda_h, tmp_u0u, linear);
      }
	for (iv=0;iv<output->wl.nlambda_h;iv++) {
	output->u0u[iv][lu][iu] = tmp_u0u[iv]*output->wl.fbeam[iv];
      }
    }

    for (j=0;j<input.rte.nphi;j++) {
      for (iu=0;iu<input.rte.maxumu;iu++) {
	for (iv=0;iv<output->wl.nlambda_t;iv++) {
	  tmp_uu_t[iv] = output->uu_t[iv][lu][j][iu];
	}
	if ( output->wl.nlambda_t == 1 ) {
	  iv = 0;
	  tmp_uu[iv]      = tmp_uu_t[iv];
	} 
	else {
	  linear = 0;
	  status = arb_wvn( output->wl.nlambda_t, output->wl.lambda_t, tmp_uu_t, 
			    output->wl.nlambda_h, output->wl.lambda_h, tmp_uu, linear);
	}
	for (iv=0;iv<output->wl.nlambda_h;iv++) {
	  output->uu[iv][lu][j][iu] = tmp_uu[iv];
	}
      }
    }

  }

  /*  free(tmp_dfdt);*/
  free(tmp_flup);
  free(tmp_rfldn);
  free(tmp_rfldir);
  free(tmp_uavg);
  free(tmp_uavgdn);
  free(tmp_uavgso);
  free(tmp_uavgup);
  free(tmp_u0u);
  free(tmp_uu);
  free(tmp_flup_t);
  free(tmp_rfldn_t);
  free(tmp_rfldir_t);
  free(tmp_uavg_t);
  free(tmp_uavgdn_t);
  free(tmp_uavgso_t);
  free(tmp_uavgup_t);
  free(tmp_u0u_t);
  free(tmp_uu_t);

  return 0;

}


